/**************************************************************/
/*    Title:    Firmware fr Auschaltverzgerung V1.1	  */
/*    Author:   Bitschnau Lukas							      */
/*    Date:     12.02.2012									  */
/*    Software: AVR-GCC / AVR Studio 4.12					  */
/*    Hardware: ATmega16							  */
/*    Note:     fclk=12MHz									  
Description:
Diese Firmware ist fr einen ATMEGA16 mit 12MHz Quarz geschrieben,
es soll dabei am Ausgang PB3 (OCR0, PWM) eine PWM ausgegeben werden, die 
es ermglicht beim schalten eines Tasters an PA0 (low aktive mit Pullup)
den PB3 verzgert herunterzudimmen. Die Verzgerungszeit ist ber
ein Potentiometer an AN5 einstellbar von 0 - 25 Sekunden. Die Dimmzeit ist minimal 5sekunden und maximal 5*255 = 1275Sekunden

Variable fr die Eistellung:
 runterdimmzeit (init) = 1 // stellt den Wert ein wielange es dauert, den Ausgang von ganz ein bis ganz dunkel werden zu lassen.


// Verbesserungen:
Es knnte zustzlich im Unterprogramm runterdimmen() die Tasten abfragen und bei der einer nderung die Dimmung 
stoppen und sofort auf PWM = 100% springen, bentige ich aber nicht.
/**************************************************************/


#include <AVR/io.h>
#include <AVR/interrupt.h>
#include <AVR/signal.h>
#include <AVR/delay.h>

unsigned char i;		// Variable fr Teiler
unsigned char k;		// Variable fr Dimm Zeit
unsigned int max = 0;	// invertrierende PWM --> max ist nicht mehr 255 sondern 0!
unsigned char enable = 0x00;
unsigned char old = 0x00;

unsigned char runterdimmzeit = 1; // Runterdimmzeit = 5Sekunden * x (Timer interupt wird alle 5sekunden ausgefhrt bei max = 0);

unsigned char teiler = 46; 	// Vorteiler fr Zeitzhler		
unsigned int time;  		// Zeitzhler in sekunden @ teiler = 46;
							// Zeitzhler in Zentelsekunden @ teiler = 5;

unsigned char ausschaltzeit; 

SIGNAL (SIG_OVERFLOW0) // Interupt fr TIMER0 (PWM), freq = CPUtakt/(256*Prescaler) = ~46Hz
{

	i++;
	k++;
	if(i == teiler)		
		{
			// Hier wird sekndlich die Variable time erhht.
			i = 0;
			time++;
			// Active LED blinkt langsam mit 1Hz --> normal mode.
			PORTB ^= (1<<PB0);
		}

	// Active led blinkt schnell (befindet sich im runterdimm modus)
	if(((i == 10)||(i ==20)||(i==30)||(i == 40)) && (old == 0x01))
		{
			PORTB ^= (1<<PB0);
		}
	

	if(k == runterdimmzeit)
	{
		k = 0;
			if(enable == 0x01)
			{
				max = max++;

			}
	}

}


void init_ports(void)
{

 	DDRD &= ~(1 << DDD2); // PD2 als eingnge definie
	DDRA &= ~(1<<DDA0); // Port PA0 als eingang fr ADC
	DDRB |= (1<<DDB3); // PWM-Port als Ausgang
//	PORTA |= (1<<PA0) | (1<<PA1); // Pullups an PA0 und PA1 aktivieren
	PORTD |= (1<<PD2); // Pullups an PA0 und PA1 aktivieren
	DDRB |= (1<<DDB0); // LED fr run anzeige

}

void adc_init(void)
{
	// AVR-Turorial: ADC Taktfrequenz sollte zwischen 50k - 200kHz liegen --> Prescaler nehmen -->
	// 12MHz / 200kHz = 60 (maximale ADCfreq.) 12MHz / 50kHz = 240 (max. ADCfreq.) --> Prescaler = 128
	ADMUX = (1<<REFS0) | (1<<ADLAR); // AVCC als Refspg, ADC0 verwendet
	ADCSRA = (1<<ADPS0) | (1<<ADPS1) | (1<<ADPS2) | (1<<ADEN) | (1<<ADSC) | (1<<ADATE); 
	//Prescaler 128, ADC Enable, AUTO Trigger enable, Automatischer Start

} 

void timer_init(void)
{	
	// PWM-Vorteiler 1024, Fast PWM, invertierend //Beachte Komentar von runterdimmen()!
	TCCR0 = (1<<WGM00) | (1<<WGM01) | (1<<COM01) | (1<<COM00) | (1<<CS02) | (1<<CS00); 
	TIMSK = (1<<TOIE0); // TimerCounter0 Overflow interup enable
	OCR0 = 255; //Initalwert setzen (PWM ist aus)

}

void runterdimmen(void)
{
	while(enable == 0x01)
	{
		PORTB |= (1<<PB0); // fr active Anzeige ( leuchtet, runterdimm mode ist aktive)
		OCR0 = max;
		if(max >= 253) // auch hier beachten, PWM ist invertiert!
		{
		OCR0 = 255;
			enable = 0x00;
			max = 0;
		}
		PORTB &= ~(1<<PB0);
	}



	// Problem:
	/*
	Wieso geht die LED nie ganz aus?

	Es ist normal, dass die LED selbst bei OCR1A = 0 immer noch ganz schwach leuchtet. 
	Die Hardware PWM funktioniert so, dass bei einem Timerwert von 0 auf jeden Fall der
	Ausgang ein geschaltet wird. Danach kommt der Compare Match bei 0 und schaltet gleich wieder aus. 
	Daher ist der Ausgang fr einen PWM-Takt eingeschaltet. 
	Um das zu ndern muss man entweder invertierte PWM nutzen, dann ist allerdings der Ausgang nie zu 100% High,
	sondern hat immer einen Takt Low beim maximalem PWM-Wert. Oder man schaltet bei 0 einfach die PWM-Funktion ab 
	und und setzt den Ausgang normal auf Low. [1]. 

	(c) by http://www.mikrocontroller.net/articles/LED-Fading

	Ich entscheide mich fr die Bentzung einer Invertierenden PWM.
*/
}
			
int main(void)
{
	init_ports();
	adc_init();
	timer_init();
	sei();

	int adc_wert_aktuell = 0; // wird der aktuelle ADC wert zwischengespeichert.

	while(1)
	{
	
		if(ADCSRA & 0x10)
			 adc_wert_aktuell = ADCH;

		ausschaltzeit = adc_wert_aktuell/10;


		 //Taster sind LOW aktive
			if (!(PIND & (1<<PIND2)))	//Taste S0 gedrckt
				{
					i = 0;
					time = 0;
					OCR0 = 0; // Invertierende PWM
					old = 0x01; // zum "merken" eines Tastenddrucks
		
				}


			// Die Timerzeit wird mit der zuvor eingestellten Zeit verglichen, ist diese
			// gleich der ausschaltzeit und der Taster zuvor gedrckt, wird die LED eingeschalten
			if((time == ausschaltzeit) && (old == 0x01))
			{	
				enable = 0x01;
				old = 0x00;
				runterdimmen();
			}

	}

}

